(library (ast-flow-graphs)
  (export
   with-AST-flow-graph
   AST-flow-graph ;; TODO: restrict this interface
   make-AST-node
   AST-flow-graph-print
#|
   ;; DEBUG ONLY
   with-local-Formal-in-type
   with-local-Formal-out-bitype
   Formal-in-type-set!
   Formal-out-bitype-set!
|#
   )
  (import (except (rnrs) case cond if) (safe-forms) (srfi-39) (flow-graphs)
          (ast) (idents) (types) (bitypes))

;; make-AST-* versions expect nodes as argument
;; AST->* versions expect ASTs as argument

#|
;; DEBUG ONLY (nodes)
(define-hashtable-parameter with-local-Formal-in-type
  Formal-in-type Formal-in-type-set!)
(define-hashtable-parameter with-local-Formal-out-bitype
  Formal-out-bitype Formal-out-bitype-set!)
|#

(define AST-flow-graph-parameter (make-parameter #f))
(define (AST-flow-graph) (AST-flow-graph-parameter))

(define-syntax with-AST-flow-graph
  (syntax-rules ()
    [(_ graph expr0 expr* ...)
     (parameterize ([AST-flow-graph-parameter graph]) expr0 expr* ...)]))

(define-syntax make-AST-node
  (syntax-rules ()
    [(_ ast make arguments ...)
     (let ([ast-var ast])
       (assert (or (AST? ast-var) (case-lambda-clause? ast-var)))
       (let ([node (make (AST-flow-graph) arguments ...)])
         (AST-contained-nodes-add! ast-var node)
         node))]))

(define (AST-contained-nodes-add! ast node)
  (cond
   [(AST? ast)
    (AST-contained-nodes-set! ast (cons node (AST-contained-nodes ast)))]
   [(case-lambda-clause? ast)
    (case-lambda-clause-contained-nodes-set!
     ast (cons node (case-lambda-clause-contained-nodes ast)))]))

(define (AST-flow-graph-print ast)
  (define (obj->string obj)
    (call-with-string-output-port (lambda (port) (write obj port))))

  (define (ident->string ident)
    (string-append "[" (obj->string (ident-name ident))
                   " " (obj->string (ident-number ident))
                   " " (obj->string (ident-assigned ident)) "]"))

  ;; contain : string * list of node * list of ast
  (define (printer contain ast)
    (cond
     [(case-lambda-clause? ast)
      (contain
       (string-append
        "(Clause "
        (let ([formals (vector->list (case-lambda-clause-formals ast))])
          (string-append
           "("
           (apply string-append (map ident->string formals))
;;             ";"
;;             (apply string-append
;;                    (map obj->string
;;                         (map flow-graph-node-output
;;                              (map Formal-in-type formals (map (lambda (x) #f)
;;                                                               formals)))))
;;             ";"
;;             (apply string-append
;;                    (map obj->string
;;                         (map flow-graph-node-output
;;                              (map Formal-out-bitype formals (map (lambda (x) #f)
;;                                                             formals)))))
           ")"))
        (let ([rest (case-lambda-clause-rest ast)])
          (if rest (ident->string rest) "#f"))
        ")")
       (case-lambda-clause-contained-nodes ast)
       (list (case-lambda-clause-body ast)))]
     [else
      (contain
       (string-append
        (AST-case ast
          [(Ref var) (string-append "(Ref " (ident->string var) ")")]
          [(Quote obj) (string-append "(Quote " (obj->string obj) ")")]
          [(Primref safe name) (string-append "(Primref " (if safe "#t" "#f")
                                              " " (symbol->string name) ")")]
          [(Set var body) (string-append "(Set " (ident->string var) ")")]
          [(App fun args) "(App)"]
          [(If test conseq altern) "(If)"]
          [(CaseLambda table clauses) (string-append "(CaseLambda "
                                                     (obj->string table) ")")]
          [(Seq e1 e2) "(Seq)"]
          [(Letrec bindings body)
           (string-append
            "(Letrec "
            (apply string-append (vector->list
                                  (vector-map
                                   ident->string (vector-map
                                                  letrec-binding-lhs bindings))))
            ")")])
        ":"
        (obj->string (flow-graph-node-output (AST-output-node ast))))
       (AST-contained-nodes ast)
       (AST-case ast
         [(Ref var) '()]
         [(Quote obj) '()]
         [(Primref safe name) '()]
         [(Set var body) (list body)]
         [(App fun args) (cons fun (vector->list args))]
         [(If test conseq altern) (list test conseq altern)]
         [(CaseLambda table clauses) (vector->list clauses)]
         [(Seq e1 e2) (list e1 e2)]
         [(Letrec bindings body)
          (append (map letrec-binding-rhs (vector->list bindings)) (list body))]))]))

  (define (labeler node)
    (string-append
     (flow-graph-node-info node) ":" (obj->string (flow-graph-node-output node))))
  (flow-graph-print printer labeler ast))

)
