(library (bitypes)
  (export
   make-bitype ;; type * type -> bitype
   bitype? ;; obj -> bool
   bitype=? ;; bitype * bitype -> bool
   bitype-true ;; bitype -> type
   bitype-false ;; bitype -> type

   bitype->type ;; bitype into join of true and false
   type->bitype ;; type into true and false
   type->bitype-true ;; type into true only
   type->split-bitype ;; true values into true and false values into false

   bitype->void-bitype ;; true and false go to both true and false
   bitype-true->void-bitype ;; only true goes to true and/or false
   bitype-false->void-bitype ;; only false goes to true and/or false

   bitype-union ;; bitype * bitype -> bitype
   bitype-intersect ;; bitype * bitype -> bitype
   )
  (import (rnrs) (types) (records-io))

(define-record-type bitype
  (nongenerative)
  (fields true false)
  (protocol
   (lambda (make-bitype)
     (lambda (true false)
       (assert (type? true))
       (assert (type? false))
       (make-bitype true false)))))

(define (bitype=? x y)
  (and (type=? (bitype-true x) (bitype-true y))
       (type=? (bitype-false x) (bitype-false y))))

(define (bitype->type bitype)
  (type-union (bitype-true bitype) (bitype-false bitype)))

(define (type->bitype type) (make-bitype type type))
(define (type->bitype-true type) (make-bitype type bot-type))

(define (type->split-bitype type)
  (make-bitype (type-intersect type true-type)
               (type-intersect type false-type)))

(define (bitype->void-bitype bitype) (type->bitype (bitype->type bitype)))

(define (bitype-true->void-bitype bitype) (type->bitype (bitype-true bitype)))
(define (bitype-false->void-bitype bitype) (type->bitype (bitype-false bitype)))

(define (bitype-union t1 t2)
  (make-bitype (type-union (bitype-true t1) (bitype-true t2))
               (type-union (bitype-false t1) (bitype-false t2))))

(define (bitype-intersect t1 t2)
  (make-bitype (type-intersect (bitype-true t1) (bitype-true t2))
               (type-intersect (bitype-false t1) (bitype-false t2))))

(record-writer (record-type-descriptor bitype)
  (lambda (r p wr)
    (display "#<" p)
    (wr (bitype-true r) p)
    (display "/" p)
    (wr (bitype-false r) p)
    (display ">" p)))

)
