(library (env-flow-functions)
  (export
   cond-env-flow-function ;; macro:
                          ;;   (_ ([test conseq ...]) ([test conseq] ...))
                          ;;   -> env-flow-function
   env-flow-function=? ;; env-flow-function * env-flow-function -> bool
   env-flow-function-zero ;; env-flow-function
   env-flow-function-identity ;; env-flow-function
   env-flow-function-compose ;; env-flow-function *
                             ;; env-flow-function ->
                             ;; env-flow-function (associative)
   env-flow-function-apply ;; env-flow-function * bitype * type -> bitype
   true false bypass ;; aux keywords
   )
  (import (rnrs) (types) (bitypes) (records-io))

(define true 0) ;; aux keyword, value doesn't matter
(define false 0) ;; aux keyword, value doesn't matter
(define bypass 0) ;; aux keyword, value doesn't matter

;; Constants for the bit level encoding
(define true-base 4)
(define false-base 0)
(define true-true-bit (+ true-base 2))
(define true-false-bit (+ true-base 1))
(define true-bypass-bit (+ true-base 0))
(define false-true-bit (+ false-base 2))
(define false-false-bit (+ false-base 1))
(define false-bypass-bit (+ false-base 0))
(define true-mask (fxior (fxarithmetic-shift 1 true-true-bit)
                           (fxarithmetic-shift 1 true-false-bit)
                           (fxarithmetic-shift 1 true-bypass-bit)))
(define false-mask (fxior (fxarithmetic-shift 1 false-true-bit)
                            (fxarithmetic-shift 1 false-false-bit)
                            (fxarithmetic-shift 1 false-bypass-bit)))
(define bypass-mask (fxior (fxarithmetic-shift 1 true-bypass-bit)
                             (fxarithmetic-shift 1 false-bypass-bit)))

(define-syntax flow-mask
  (syntax-rules (true false bypass)
    [(_ true true) (fxarithmetic-shift 1 true-true-bit)]
    [(_ true false) (fxarithmetic-shift 1 true-false-bit)]
    [(_ true bypass) (fxarithmetic-shift 1 true-bypass-bit)]
    [(_ false true) (fxarithmetic-shift 1 false-true-bit)]
    [(_ false false) (fxarithmetic-shift 1 false-false-bit)]
    [(_ false bypass) (fxarithmetic-shift 1 false-bypass-bit)]))

(define-syntax flow-masks
  (syntax-rules ()
    [(_ _) 0]
    [(_ out in) (flow-mask out in)]
    [(_ out in ins ...) (fxior (flow-mask out in) (flow-masks out ins ...))]))

(define-syntax $make-env-flow-source
  (syntax-rules ()
    [(_ _) 0]
    [(_ out [test conseq]) (if test (flow-masks out conseq) 0)]
    [(_ out [test conseq] . rest)
     (fxior (if test (flow-masks out conseq) 0)
            ($make-env-flow-source out . rest))]))

(define-syntax cond-env-flow-function
  (syntax-rules ()
    [(_ ([true-test true-conseq] ...) ([false-test false-conseq] ...))
     (fxior
      ($make-env-flow-source true [true-test true-conseq] ...)
      ($make-env-flow-source false [false-test false-conseq] ...))]))

(define env-flow-function-zero 0)

(define env-flow-function-identity
  (fxior (flow-mask true true) (flow-mask false false)))

;; env-flow-function * env-flow-function -> bool
(define (env-flow-function=? x y) (fx=? x y))

;; env-flow-function * env-flow-function -> env-flow-function
;; f * g -> (f.g)
(define (env-flow-function-compose outer inner)
  (fxior
   (fxior ;; WORKAROUND: Chez does strange allocation if we don't do extra fxior
    (fxand
     inner
     (if (fxbit-set? outer true-true-bit)
         (if (fxbit-set? outer false-false-bit)
             (fxior true-mask false-mask) true-mask)
         (if (fxbit-set? outer false-false-bit)
             false-mask 0)))
    (if (fxbit-set? outer true-false-bit)
        (fxand true-mask (fxarithmetic-shift inner (- true-base false-base)))
        0)
    (if (fxbit-set? outer false-true-bit)
        (fxand false-mask (fxarithmetic-shift inner (- false-base true-base)))
        0))
   (fxand bypass-mask outer)))

;; env-flow-function * bitype * type -> bitype
(define (env-flow-function-apply function bitype bypass)
  (define (env-flow-source-apply source)
    (define t1 (if (fxbit-set? source true-true-bit)
                   (bitype-true bitype)
                   bot-type))
    (define t2 (if (fxbit-set? source true-false-bit)
                   (type-union t1 (bitype-false bitype))
                   t1))
    (define t3 (if (fxbit-set? source true-bypass-bit)
                   (type-union t2 bypass)
                   t2))
    t3)
  (assert (bitype? bitype))
  (assert (type? bypass))
        
  (make-bitype
   (env-flow-source-apply function)
   (env-flow-source-apply (fxarithmetic-shift function (- true-base false-base)))))

)
