(library (env-flow-graphs)
  (export
   make-env-flow-graph ;; ast -> void
   env-flow-graph-force! ;; ast -> void
   )
  (import (except (rnrs) case cond if) (safe-forms) (composition-stacks)
          (env-flow-functions) (ast) (ast-flow-graphs) (types) (flow-graphs))

;;;;;;;;;;;;;;;;
;; defines transfer function from output of child to output of this
;; thus does one recur per child

(define (make-env-flow-graph ast)
  ;; context :: composition-stack of flow-graph-node of env-flow-function
  (define (rec ast context)

    (define-syntax define-rec
      (syntax-rules ()
        [(_ (name args ...)
            ([true-test true-conseq] ...)
            ([false-test false-conseq] ...))
         (define (name ast context expr args ...)
           (define node
             (make-AST-node ast make-flow-graph-lazy-node (symbol->string 'name)
                            env-flow-function=? env-flow-function-zero))
           (AST-graph-context-thunk-set!
            expr (lambda ()
                   (flow-graph-lazy-node-force!
                    node
                    (lambda (args ...)
                      (cond-env-flow-function
                       ([true-test true-conseq] ...)
                       ([false-test false-conseq] ...)))
                    (AST-output-node args) ...)))
           (rec expr (composition-stack-cons node context)))]))

    (define-rec (rec-id)
      ([#t true])
      ([#t false]))

    (define-rec (rec-bypass)
      ([#t bypass])
      ([#t bypass]))

    (define-rec (rec-void ast)
      ([(true-type? ast) true]  [(true-type? ast) false])
      ([(false-type? ast) true] [(false-type? ast) false]))

    (define-rec (rec-if-test conseq altern)
      ([(true-type? conseq) true]  [(true-type? altern) false])
      ([(false-type? conseq) true] [(false-type? altern) false]))

    (define-rec (rec-if-branch this that)
      ([#t true]  [(true-type? that) bypass])
      ([#t false] [(false-type? that) bypass]))

    (AST-graph-context-set! ast context)

    (AST-case ast
      [(Ref var) (values)]
      [(Quote obj) (values)]
      [(Primref safe name) (values)]
      [(Set var expr) (rec-void ast context expr ast)]

      [(App fun actuals)
       (rec-void ast context fun ast)
       (vector-for-each (lambda (actual) (rec-void ast context actual ast))
                        actuals)]

      [(If test conseq altern)
       (rec-if-test ast context test conseq altern)
       (rec-if-branch ast context conseq conseq altern)
       (rec-if-branch ast context altern altern conseq)]

      [(CaseLambda table clauses)
       (vector-for-each
        (lambda (clause)
          ;; NOTE: context here is only used for in not out
          (rec-bypass ast context (case-lambda-clause-body clause)))
        clauses)]

      [(Seq expr1 expr2)
       (rec-void ast context expr1 ast)
       (rec-id ast context expr2)]

      [(Letrec bindings body)
       (vector-for-each
        (lambda (binding) (rec-void ast context (letrec-binding-rhs binding) ast))
        bindings)
       (rec-id ast context body)]))

  (rec ast (make-composition-stack-null
            (lambda (inner outer)
              (make-AST-node ast make-flow-graph-function-node
                             "env-flow-function-compose"
                             env-flow-function=?
                             env-flow-function-compose outer inner)))))

(define (env-flow-graph-force! ast)
  (define (rec ast) (env-flow-graph-force! ast))
  (let ([thunk (AST-graph-context-thunk ast)])
    (when thunk
      (thunk)
      (AST-graph-context-thunk-set! ast #f)))

  (AST-case ast
    [(Ref var) (values)]
    [(Quote obj) (values)]
    [(Primref safe name) (values)]
    [(Set var expr) (rec expr)]
    [(App fun actuals)
     (rec fun)
     (vector-for-each rec actuals)]
    [(If test conseq altern)
     (rec test)
     (rec conseq)
     (rec altern)]
    [(CaseLambda table clauses)
     (vector-for-each (lambda (clause) (rec (case-lambda-clause-body clause)))
                      clauses)]
    [(Seq expr1 expr2)
     (rec expr1)
     (rec expr2)]
    [(Letrec bindings body)
     (vector-for-each (lambda (binding) (rec (letrec-binding-rhs binding)))
                      bindings)
     (rec body)]))

)
