(library (safe-forms)
  (export if cond case)
  (import (rename (rnrs) (if rnrs:if) (cond rnrs:cond) (case rnrs:case)))

;; Versions of 'if', 'cond', and 'case' that are slightly safer to use:
;;   'if' - rejects one-armed if
;;   'cond' - errors if no cases match
;;   'case' - errors if no cases match, also allows a bare symbol
;; There are modified from R6RS - Appendix B

(define-syntax if
  (lambda (x)
    (syntax-case x ()
      [(_ t c) (raise (syntax-violation 'if "found one-armed if" x))]
      [(_ t c a) #'(rnrs:if t c a)])))

(define-syntax cond
  (syntax-rules (else =>)
    ((cond (else result1 result2 ...))
     (begin result1 result2 ...))
    ((cond (test => result))
     (let ((temp test))
       (if temp (result temp) (assertion-violation 'cond "no clauses match"))))
    ((cond (test => result) clause1 clause2 ...)
     (let ((temp test))
       (if temp
           (result temp)
           (cond clause1 clause2 ...))))
    ((cond (test)) (or test (assertion-violation 'cond "no clauses match")))
    ((cond (test) clause1 clause2 ...)
     (let ((temp test))
       (if temp
           temp
           (cond clause1 clause2 ...))))
    ((cond (test result1 result2 ...))
     (if test (begin result1 result2 ...)
         (assertion-violation 'cond "no clauses match")))
    ((cond (test result1 result2 ...)
           clause1 clause2 ...)
     (if test
         (begin result1 result2 ...)
         (cond clause1 clause2 ...)))))

(define-syntax case
  (syntax-rules (else)
    [(case e
       [s* r* r** ...]
       ...
       [else else-r else-r* ...])
     (let ([t e])
       (cond
         [(case-test t s*) r* r** ...]
         ...
         [else else-r else-r* ...]))]
    [(case e
       [s r1 r1* ...]
       [s* r2* r2** ...]
       ...)
     (let ([t e])
       (cond
         [(case-test t s) r1 r1* ...]
         [(case-test t s*) r2* r2** ...]
         ...
         [else (assertion-violation 'case "no clauses match")]))]))

(define-syntax case-test
  (syntax-rules ()
    [(_ v (m* ...)) (memv v '(m* ...))]
    [(_ v m) (eqv? v 'm)]))

)
