(library (srfi-39)
  (export make-parameter parameterize)
  (import (rnrs))

(define make-parameter
  (case-lambda
   [(init) (make-parameter init (lambda (x) x))]
   [(init converter)
    ;; TODO: We simulate a parameter here.  For thread safety this
    ;; should become define-threaded which uses a threaded-parameter.
    (let ([value (converter init)])
      (let ([parameter ;; give the lambda a name
             (case-lambda
              [() value]
              [(new-value) (set! value (converter new-value))])])
        parameter))]))

;; WARNING: converters must be idempotent
(define-syntax parameterize
  (lambda (stx)
    (syntax-case stx ()
      [(_ ([lhs rhs] ...) body body* ...)
       (with-syntax ([(param ...) (generate-temporaries #'(lhs ...))]
                     [(local ...) (generate-temporaries #'(rhs ...))])
         #'(let ([param lhs] ...
                 [local rhs] ...)
             (let ([swap (lambda ()
                           (let ([tmp (param)])
                             (param local)
                             (set! local tmp)) ...)])
               (dynamic-wind swap (lambda () body body* ...) swap))))])))

)
