(library (syntax-helpers)
  (export
   syntax-ellipsis? ;; syntax -> bool
   syntax-indexes ;; #'(a b c ...) -> #'((1 . a) (2 . b) (3 . c) ...)
   syntax-length ;; syntax -> int
   syntax-car syntax-cdr ;; syntax -> syntax
   syntax-map syntax-for-all
   syntax-append)
  (import (rnrs))

(define (syntax-ellipsis? stx)
  (and (identifier? stx) (free-identifier=? stx #'(... ...))))

(define syntax-indexes
  (case-lambda
   [(list) (syntax-indexes list 0)]
   [(list base)
    (syntax-case list ()
      [() #'()]
      [(x . y) (with-syntax ([base base]
                             [y (syntax-indexes #'y (+ base 1))])
                 #'(base . y))])]))

(define (syntax-length stx)
  (syntax-case stx ()
    [() 0]
    [(car . cdr) (+ 1 (syntax-length #'cdr))]))

(define (syntax-car stx) (syntax-case stx () [(car . cdr) #'car]))
(define (syntax-cdr stx) (syntax-case stx () [(car . cdr) #'cdr]))

(define (syntax-map function . stx)
  (syntax-case stx ()
    [(() ...) #'()]
    [else
     #`(#,(apply function (map syntax-car stx)) .
        #,(apply syntax-map function (map syntax-cdr stx)))]))

(define (syntax-for-all function . stx)
  (syntax-case stx ()
    [(() ...) #t]
    [else
     (and (apply function (map syntax-car stx))
          (apply syntax-for-all function (map syntax-cdr stx)))]))

(define (syntax-append template-identifier . args)
  (datum->syntax
   template-identifier
   (string->symbol
    (apply string-append
           (map (lambda (x)
                  (cond
                   [(string? x) x]
                   [(symbol? x) (symbol->string x)]
                   [(identifier? x) (symbol->string (syntax->datum x))]))
                args)))))

)
