(library (tree-locations)
  (export
   make-tree-location-table
   tree-location-table-key-list
   tree-location-table-prev-ref
   tree-location-table-next-ref)
  (import (rnrs))

(define-record-type
  (tree-location-table $make-tree-location-table tree-location-table?)
  (nongenerative)
  (fields
   key-locations ;; key -> vector[node]
   key-nodes ;; key -> vector[node]
   key-list ;; list[key]
   ))

;; walker must call (for-each contains children) and (mark key)

;; tree-location ::
;;   obj * (contains * mark * obj -> void) -> tree-location-table
;; mark :: key -> void
;; contains :: obj -> void

(define-syntax make-tree-location-table
  (syntax-rules ()
    [(_ tree ;; node
        node-start-location-set! ;; node * int -> void
        node-end-location-set! ;; node * int -> void
        key-nodes ;; key -> list or vector of node
        key-nodes-set! ;; key * list or vector of node -> void
        key-locations ;; key -> list or vector of int
        key-locations-set! ;; key * list or vector of int -> void
        contains mark add
        walker-fun)

     (let ([counter 0]
           [key-list '()])
       (letrec ([walker walker-fun]
                [contains (lambda (node)
                            (node-start-location-set! node counter)
                            (set! counter (+ 1 counter))
                            (walker node)
                            (set! counter (+ 1 counter))
                            (node-end-location-set! node counter))]
                [mark
                 (lambda (node key)
                   (key-nodes-set!     key (cons node    (key-nodes key)))
                   (key-locations-set! key (cons counter (key-locations key))))]
                [add (lambda (key) (set! key-list (cons key key-list)))])

         ;; walk the tree
         (contains tree)

         ;; Reverse and convert to vectors so we can vector-binary-search
         (for-each
          (lambda (key)
            (key-nodes-set!     key (list->vector (reverse (key-nodes     key))))
            (key-locations-set! key (list->vector (reverse (key-locations key)))))
          key-list)

         ($make-tree-location-table key-locations key-nodes key-list)))]))

(define (key-locations<=? ref key) (<= ref key))

;; returns index of first greater or equal
(define vector-binary-search
  (case-lambda
   [(vector key) (vector-binary-search vector key 0 (vector-length vector))]
   [(vector key start len)
    (define rec vector-binary-search)
    (let* ([len1 (div len 2)]
           [mid (+ start len1)]
           [len2 (- len len1 1)])
      (cond
       [(= len 0) start]
       [(key-locations<=? key (vector-ref vector mid)) (rec vector key start len1)]
       [else (rec vector key (+ mid 1) len2)]))]))

(define (tree-location-table-prev-ref tpt node key default start-location
                                      key-locations key-nodes)
  (define index (vector-binary-search key-locations start-location))
  (if (= index 0)
      default
      (vector-ref key-nodes (- index 1))))

(define (tree-location-table-next-ref tpt node key default end-location
                                      key-locations key-nodes)
  (define index (vector-binary-search key-locations end-location))
  (if (= index (vector-length key-locations))
      default
      (vector-ref key-nodes index)))

)
